﻿using JudgeModule;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public class JudgeManager : MonoBehaviour
{
    float latency = 225f;
    float scrollMultiplier = 1.0f;
    float startOffset = -5f;
    float baseTime;
    bool IsInitTime;

    bool IsGameEnd;

    private GameObject offset;
    private GameObject noteobj;
    private GameObject deactives;
    private GameObject appear;
    private GameObject disappear;
    private MotionGageManager motionGageManager;
    private GameObject judgeText;

    private GameObject motionSampleDisplayPrefab;

    private Vector3 initialPos;

    private float MsPerBeat
    {
        get
        {
            return 60 * 1000f / GameManager.Instance.CurrentTrack.BPM;
        }
    }
    
    private float ScrollSpeed
    {
        get
        {
            return scrollMultiplier * MsPerBeat / 1000f;
        }
    }

    private InputManager input;
    private NoteCondition condition;
    private List<Note> notes;
    private NoteJudger judger;

    void SetObjectRef()
    {
        offset = GameObject.Find("Offset");
        noteobj = GameObject.Find("Noteobj");
        deactives = GameObject.Find("Deactives");
        appear = GameObject.Find("NoteAppear");
        disappear = GameObject.Find("NoteDisappear");
        judgeText = GameObject.Find("Judge");
        input = GameObject.Find("InGameManagers").GetComponent<InputManager>();
        condition = new NoteCondition(input);
        judger = new NoteJudger(new Dictionary<string, GameObject> { }, ShowResult);
        motionGageManager = new MotionGageManager
        { motionGuage = GameObject.Find("Motion Guage") };
    }

    void JudgeInit()
    {
        initialPos = offset.transform.position;
        judgeText.SetActive(false);
        motionGageManager.ResetGuage();
        IsGameEnd = false;
        baseTime = 0f;
        IsInitTime = false;
    }

    void LoadGameObjects()
    {
        motionSampleDisplayPrefab = Resources.Load("Motion Sample Display") as GameObject;
        notes = GameManager.Instance.CurrentTrack.Notes.ToList();
        notes.ForEach(LoadNote);

        var sounds = GameObject.Find("Sounds");
        sounds.transform.Cast<Transform>()
                        .ToList()
                        .ForEach(x => x.gameObject.SetActive(false));
        if (GameManager.Instance.CurrentTrack.BGM != null)
            sounds.transform.Find(GameManager.Instance.CurrentTrack.BGM)
                  .gameObject.SetActive(true);
    }
    
    // Use this for initialization
    void Start()
    {
        SetObjectRef();
        JudgeInit();
        LoadGameObjects();
    }

    // Update is called once per frame
    void Update()
    {
        float current = motionGageManager.UpdateTime();
        if (!IsInitTime)
        {
            baseTime = current;
            IsInitTime = true;
        }

        float timing = current - baseTime;

        offset.transform.position = Vector3.left * timing * ScrollSpeed;
        
        motionGageManager.UpdateGuage();

        if (noteobj.transform.childCount == 0 ||
            IsGameEnd)
            Invoke("ShowResult", 2f);

        RelocateControllers(timing);

        JudgeNote(timing);
    }

    void RelocateControllers(float timing)
    {
        DeactiveUndecided();
        SwitchStandBy();
        AssignNotes(timing);
    }

    void DeactiveUndecided()
    {
        noteobj.transform
               .Cast<Transform>()
               .Where(c => initialPos.x - c.position.x > Judge.BAD.ButtonTimingRange)
               .ToList()
               .ForEach(x => x.SetParent(deactives.transform));
    }

    void SwitchStandBy()
    {
        deactives.transform
                 .Cast<Transform>()
                 .Select(x => x.GetComponent<Controller>())
                 .Where(c => c.transform.position.x < disappear.transform.position.x)
                 .ToList()
                 .ForEach(x =>
                 {
                     x.transform.position = appear.transform.position;
                     x.transform.SetParent(appear.transform);
                     x.Instance = null;
                     x.gameObject.SetActive(false);
                 });
    }

    void AssignNotes(float timing)
    {
        appear.transform
              .Cast<Transform>()
              .Select(x => x.GetComponent<Controller>())
              .ToList()
              .ForEach(c =>
              {
                  c.Instance = GetLastNote(timing);
                  if (c.Instance != null)
                  {
                      c.Instance.Component = c;
                      c.transform.position = initialPos + Vector3.right * ((c.Instance.StartTiming - timing) * ScrollSpeed);
                      c.transform.SetParent(noteobj.transform);
                      c.gameObject.SetActive(true);
                  }
              });
    }

    void JudgeNote(float timing)
    {
        Note note = GetLastNote(timing);
        if (note == null)
            return;

        if (condition.IsWrongInput(note))
        {
            judger.WrongNoteProc(note);
            return;
        }

        if (Judge.IsNoteStart(note, timing))
            PlaySampleMotion(note as MotionNote);

        if (note.IsLong)
            JudgeLongNote(note, timing);
        else
            JudgeShortNote(note, timing);
    }

    Note GetLastNote(float timing)
    {
        while (noteobj.transform.childCount != 0)
        {
            var note = noteobj.transform.GetChild(0).gameObject
                              .GetComponent<Controller>().Instance;
            if (!Judge.IsPastNote(note, timing))
                return note;
            judger.SetJudge(Judge.MISS);
            judger.DeactivateNote(note);
        }

        return null;
    }

    void PlaySampleMotion(MotionNote note)
    {
        GameObject motionSample = Instantiate(motionSampleDisplayPrefab);
        MotionSampleDisplay msd = motionSample.GetComponent<MotionSampleDisplay>();
        msd.sprite = note.Image;
        msd.timeout = MsPerBeat;
        note.Activated = true;
        note.MotionSampleDisplay = msd;
    }

    void JudgeShortNote(Note note, float timing)
    {
        if (condition.IsShortNoteEntered(note))
            judger.EnteredNoteProc(note, timing);
    }

    void JudgeLongNote(Note note, float timing)
    {
        if      (condition.IsLongNoteStartCorrectly(note))
            judger.EnteredNoteProc(note, timing);
        else if (condition.IsLongNoteFinishCorrectly(note, timing))
            judger.CorrectlyStoppedNoteProc(note, timing);
        else if (condition.IsLongNoteFinishIncorrectly(note, timing))
            judger.IncorrectlyStoppedNoteProc(note, timing);
        else if (condition.IsLongNoteHoldCorrectly(note))
            judger.ContinuingNoteProc(note, timing);
    }

    void LoadNote(Note note)
    {
        GameObject obj = Instantiate(
                Resources.Load(note.Type.ToString(), typeof(GameObject)),
                noteobj.transform)
                as GameObject;

        var controller = obj.AddComponent<Controller>();
        controller.Instance = note;
        note.Component = controller;

        if (note.Type == NoteType.MeasureLine || note.Type == NoteType.BeatLine)
            obj.transform.SetParent(offset.transform);

        if (note.IsLong)
            StretchLongNote(note, obj);

        obj.transform.position += initialPos +
            (Vector3.right * (note.StartTiming * ScrollSpeed));
    }

    void StretchLongNote(Note note, GameObject obj)
    {
        float length = note.Length * ScrollSpeed;

        var holdTransform = obj.transform.GetChild(1)
            .GetComponent<RectTransform>();
        var endTransform = obj.transform.GetChild(2)
            .GetComponent<RectTransform>();

        holdTransform.SetSizeWithCurrentAnchors(
            RectTransform.Axis.Horizontal, length);
        holdTransform.position += (Vector3.right * length / 2);
        endTransform.position += (Vector3.right * length);
    }

    private bool onResult = false;

    private void ShowResult()
    {
        if (!onResult)
        {
            Destroy(GetComponent<InputManager>());
            Destroy(GetComponent<MotionView>());
            Destroy(GetComponent<CoordinateMapperManager>());
            Destroy(this);

            GameManager.Instance.SceneTransition("Result");
            onResult = true;
        }
    }
}
